/*
   polys.p

   Ian Wright 9.5.95

   Routines for collision detection of polygons defined
   by their vertices.

*/

uses rc_graphic;

vars procedure collision_detect get_distance return_intersections
               get_bounded_intersection in_box get_intersection
               make_polygon generate_equations get_line;

/*
         CONTENTS - (Use <ENTER> gg to access required sections)

 define :class Line;
 define :class poly;
 define :method collision_detect( x1, y1, x2, y2, polygon:poly )
 define procedure get_distance( pos1, pos2 ) -> dist;
 define :method return_intersections( x1, y1, x2, y2, polygon:poly ) -> int_points;
 define :method get_bounded_intersection(
 define procedure in_box(x1, y1, x2, y2, x, y) -> bool;
 define :method get_intersection( line1:Line, line2:Line ) -> (x, y);
 define procedure make_polygon(vertices) -> polygon;
 define :method generate_equations( polygon:poly );
 define procedure get_line( x1, y1, x2, y2 ) -> line_eq;

*/

define :class Line;

   slot Line_type; ;;; y (i.e., y = c) x (i.e., x = c) or n (y = mx + c)
   slot Line_grad;
   slot Line_cnst;

enddefine;

define :class poly;

   slot poly_points = [];
   slot poly_equations = [];

enddefine;

;;; COLLISION_DETECT
;;; Given two points (original point and point to move to) and
;;; a polygon, return true if collision occurs (false otherwise)
;;; and new point.
;;; Doesn't really need two results. Could return false or point.
;;; Should chane to return a pair?
define :method collision_detect( x1, y1, x2, y2, polygon:poly )
		-> hit -> newpos;

lvars , intersect_points, nearest_point, smallest_dist, dist, a_point;

   ;;; get all bounded intersection points
   return_intersections(x1, y1, x2, y2, polygon) -> intersect_points;
   unless null(intersect_points) then
      ;;; find the nearest point
      hd(intersect_points) -> nearest_point;
      sim_distance_from(x1, y1, explode(nearest_point)) -> smallest_dist;
      for a_point in intersect_points do
         sim_distance_from(x1, y1, explode(a_point)) -> dist;
         if dist < smallest_dist then
            a_point -> nearest_point;
            dist -> smallest_dist;
         endif;
      endfor;
      true -> hit;
      nearest_point -> newpos;
   else
      false -> hit;
      [^x2 ^y2] -> newpos;
   endunless;

enddefine;

;;; GET_DISTANCE (Not needed any more ???)
;;; Returns distance between two points
define procedure get_distance( pos1, pos2 ) -> dist;

	lvars pos1, pos2, dist;
	sim_distance_from(destpair(pos1), destpair(pos2)) -> dist;

enddefine;

;;; RETURN_INTERSECTIONS
;;; Given two points and a polygon returns all the intersection points
;;; between a line draw between the two points and the sides of the polygon.
;;; In what format? Are points pairs, or two element lists???
;;; Should changed to return pairs???
define :method return_intersections( x1, y1, x2, y2, polygon:poly ) -> int_points;

lvars x1, y1, x2, y2, polygon, int_points, movement_eq, count,
	x,y
	points = polygon.poly_points,
	eqns = polygon.poly_equations;

   get_line(x1, y1, x2, y2) -> movement_eq;
   [%
      for count from 1 to length(eqns) do
         unless count == length(eqns) then
            get_bounded_intersection(
				movement_eq,
				(eqns)(count),
               	x1, y1, x2, y2,
				explode((points)(count)), explode((points)(count + 1)))
					-> (x, y);
         else
            get_bounded_intersection(
				movement_eq,
				(eqns)(count),
        		x1, y1, x2, y2,
				explode((points)(count)), explode((points)(1))) -> (x, y);
         endunless;
         if x then
            [^x ^y]
         endif;
      endfor;
   %] -> int_points;

enddefine;

;;; GET_BOUNDED_INTERSECTION
;;; Given two bounded line equations returns intersect point (if no point
;;; returns false)
define :method get_bounded_intersection(
		line1:Line, line2:Line,
		l1x1, l1y1, l1x2, l1y2,
		l2x1, l2y1, l2x2, l2y2 ) -> (x, y);

lvars x, y;

   get_intersection(line1, line2) -> (x, y);
   ;;; intersect_point must be in both bounding boxes
   if x then
      unless in_box(l1x1, l1y1, l1x2, l1y2, x, y) and
            in_box(l2x1, l2y1, l2x2, l2y2, x, y)
	  then
         false -> x;
      endunless;
   endif;

enddefine;

;;; IN_BOX
;;; Given two coordinate bounds and a point returns true if point
;;; is within the bounding box defined by the coordinates.
define procedure in_box(x1, y1, x2, y2, x, y) -> bool;
	;;; is x,y in the box defined by the other two points

	min(x1, x2) <= x and x <= max(x1, x2)
	and
	min(y1, y2) <= y and y <= max(y1, y2) -> bool

enddefine;

;;; GET_INTERSECTION
;;; Given two line equations returns intersect point (if any)
define :method get_intersection( line1:Line, line2:Line ) -> (x, y);

lvars line1, line2, line3, x = false, y = false;

	;;; why use strings ???
   if line1.Line_type = 'n' and line2.Line_type /= 'n' then
	 ;;; why make copies?
      copy(line1) -> line3;
      copy(line2) -> line1;
      line3 -> line2;
   endif;
   false ->>(x, y);
   if line1.Line_type = 'n' and line2.Line_type = 'n' then
      ;;; two normal line equations
      if line1.Line_grad = line2.Line_grad then
         ;;; leave intersect_point null
      else
         (line2.Line_cnst - line1.Line_cnst) /
			(line1.Line_grad - line2.Line_grad) -> x;
         (line2.Line_grad * line1.Line_cnst
			- line1.Line_grad * line2.Line_cnst) /
				(line2.Line_grad - line1.Line_grad) -> y;
      endif;
   elseif line1.Line_type = 'x' then
      ;;; line1 is x = c
      switchon line2.Line_type
         case = 'n' then
            line1.Line_cnst -> x;
            line1.Line_cnst * line2.Line_grad + line2.Line_cnst -> y;
         case = 'x' then
            ;;; leave intersect_point null
         case = 'y' then
            line1.Line_cnst -> x;
            line2.Line_cnst -> y;
      endswitchon;
   elseif line1.Line_type = 'y' then
      ;;; line1 is y = c
      switchon line2.Line_type
         case = 'n' then
            (line1.Line_cnst - line2.Line_cnst) / line2.Line_grad -> x;
            line1.Line_cnst -> y;
         case = 'x' then
            line2.Line_cnst -> x;
            line1.Line_cnst -> y;
         case = 'y' then
            ;;; leave intersect_point null
      endswitchon;
   endif;

enddefine;

;;; MAKE_POLYGON
;;; Given a list of vertices, constructs a polygon.
define procedure make_polygon(vertices) -> polygon;

lvars vertices, polygon, a_point;

   newpoly() -> polygon;
   [%
      for a_point in vertices do
		;;; this just copies the point????
         [ %explode(a_point)% ];
      endfor;
   %] -> polygon.poly_points;
   generate_equations(polygon);

enddefine;

/*
    make_polygon([[10 20] [100 40] [30 200]]) =>
*/

;;; GENERATE_EQUATIONS
;;; Given a polygon with vertices defined generates the
;;; line equations corresponding to its sides.
;;; Each vertex defined by a list of two coordinates ???
define :method generate_equations( polygon:poly );

lvars polygon, num_points, point_num,
	points = polygon.poly_points;

   length(points) -> num_points;
   [%
      for point_num from 1 to num_points do
         unless point_num == num_points then
            get_line(
				explode(points(point_num)), explode(points(point_num + 1)))
         else
            get_line(explode(points(point_num)), explode(points(1)))
         endunless;
      endfor;
   %] -> polygon.poly_equations;

enddefine;

;;; GET_LINE
;;; Given two points returns the line equation running through them.
define procedure get_line( x1, y1, x2, y2 ) -> line_eq;

lvars numer, denom;

   newLine() -> line_eq;
   x2 - x1 -> denom;
   y2 - y1 -> numer;
   if denom = 0 then
      'x' -> line_eq.Line_type;
      x1 -> line_eq.Line_cnst;
   elseif numer = 0 then
      'y' -> line_eq.Line_type;
      y1 -> line_eq.Line_cnst;
   else
      'n' -> line_eq.Line_type;
      numer / denom -> line_eq.Line_grad;
      y1 - (line_eq.Line_grad * x1) -> line_eq.Line_cnst;
   endif;

enddefine;
